#   Copyright 2014 Telenor Digital AS
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import os
import sys
import transmute.bootstrap

class Transmuter(object):
    """Manage updates to Python's module search path."""

    def __init__(self, working_set):
        self.working_set = working_set

    @classmethod
    def _get_dist_conflicts(self, dist):
        conflicts = []
        for module in dist._get_metadata('top_level.txt'):
            if module == 'pkg_resources':
                continue
            if module in sys.modules:
                conflicts.append(module)
        return conflicts

    def get_conflicts(self):
        conflicts = {}
        for dist in self.working_set:
            dist_conflicts = self._get_dist_conflicts(dist)
            if dist_conflicts:
                conflicts[dist.project_name] = dist_conflicts
        return conflicts

    def _reset_path(self):
        transmute.bootstrap.reset_system_path(self.working_set)

    def soft_transmute(self):
        for dist in self.working_set:
            dist.activate()
        self._reset_path()

    def hard_transmute(self):
        self.executable = sys.executable
        self.arguments = [ self.executable ] + sys.argv
        self.environment = os.environ

        # Reset PYTHONPATH
        self._reset_path()
        os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)

        os.execve(self.executable, self.arguments, self.environment)
        assert False

    def transmute(self):
        """Inject packages in working set."""

        # TODO: Recover from a bad update
        if self.get_conflicts():
            self.hard_transmute()
        self.soft_transmute()



#!/usr/bin/env python
#
#   Copyright 2014 Telenor Digital AS
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Bootstrap installation of transmute package.

When imported (or directly executed), this module will make the latest transmute
package available, downloading it from PyPI as needed. Dowloaded packages are
stored in a local cache and will be reused in subsequent runs.

This can also be used to bootstrap applications that use the transmute package.
In this case, the module becomes the application's main executable. A couple of
customization points are provided to support this use case with minimal effort:

    requirements: global variable listing packages to be fetched from PyPI.
    main(): placeholder for application specific logic. If the module is
        executed as __main__ script, this gets called after packages in
        requirements have been updated and added to sys.path.

Additionally, bootstrap_starting(), bootstrap_succeeded() and bootstrap_failed()
are called at specific points in the bootstrapping process.

The following example could be used to load latest 'foobar' from PyPI and invoke
the foobar.cli.main() entry point:

    requirements = [ 'foobar' ]
    def main():
        import foobar.cli
        return foobar.cli.main()

"""


################################################################################
# Customization points to support running module as main application script.
################################################################################

requirements = [ 'transmute' ]

def main():
    """Called when module is '__main__', after successful bootstrap."""
    ### ~~~ Your code here ~~~ ###

def bootstrap_starting():
    """Called before attempting to load/download packages."""

def bootstrap_succeeded():
    """Called after (updated) packages have been added to sys.path."""
    global _bootstrapped_by_transmute
    _bootstrapped_by_transmute = True

def bootstrap_failed():
    """Called when bootstrap fails to download packages from PyPI or load them
    from the local cache.
    """
    raise RuntimeError("Unable to load 'transmute' package.")


################################################################################
# transmute.bootstrap code follows.
################################################################################

import os
import os.path
import sys

def _chunk_read(file, chunk_size=16*1024):
    """Read file one chunk at a time."""

    while True:
        chunk = file.read(chunk_size)
        if chunk == '':
            break
        yield chunk

def _md5(filename):
    """Compute MD5 hash of a file."""

    import hashlib

    h = hashlib.md5()
    try:
        with open(filename) as file:
            for chunk in _chunk_read(file):
                h.update(chunk)
    except: pass

    return h.hexdigest()

def _copy(source, destination, chunk_size):
    """Read source into destination, one chunk at a time."""

    for chunk in _chunk_read(source, chunk_size):
        destination.write(chunk)

def _download(source, filename, md5sum):
    """Copy source to filename, verify MD5 hash of content.

    Content is initially saved to a temporary file and the content's MD5 hash is
    verified before the file is atomically renamed to the desired filename.

    This will call source.close().
    """
    import contextlib
    import tempfile

    dirname = os.path.dirname(filename)
    dst = tempfile.NamedTemporaryFile(suffix='.download', dir=dirname)
    with contextlib.closing(dst):
        _copy(source, dst, 4 * 1024)
        dst.flush()

        if _md5(dst.name) != md5sum:
            raise RuntimeError(
                    "MD5 hash of local file doesn't match expected value")

        # Depend on non-documented implementation of NamedTemporaryFile, a
        # shame, but prettier than the alternative
        dst.delete = False

        os.rename(dst.name, filename)

def reset_system_path(working_set):
    """Prepend entries in working_set to sys.path."""

    for entry in sys.path:
        if entry not in working_set.entries:
            working_set.add_entry(entry)

    sys.path[:] = working_set.entries

    if 'pkg_resources' in sys.modules:
    # pkg_resource's global working_set may be outdated
        reload(sys.modules['pkg_resources'])

def require(working_set, baskets, *requirements):
    """Satisfy requirements from given baskets."""

    import pkg_resources
    import zipimport

    requirements = list(pkg_resources.parse_requirements(requirements))

    environment = pkg_resources.Environment()
    for basket in baskets:
        basket.fill_environment(environment, requirements)

    # Download needed distributions
    while True:
        needed = working_set.resolve(requirements, env=environment)
        for dist in needed:
            if hasattr(dist, '_transmute_basket'):
                try: dist._transmute_basket.make_local(dist)
                except: # Drop dist, start over
                    environment.remove(dist)
                    break
                dist._provider = pkg_resources.EggMetadata(
                        zipimport.zipimporter(dist.location))
        else:
            break

    for dist in needed:
        working_set.add(dist)


class Basket(object):
    """A container for Python Eggs."""

    _cache_dir = os.path.expanduser('~/.python-transmute/cache')

    def __init__(self, url=None, path=None):
        assert (path is None) != (url is None)

        if path is None:
            path = self._prepare_cache(url)

        self.url = url
        self.path = os.path.join(path, '') # Keep trailing separator!
        self._projects = set()
        self.distributions = {}

        try:
            for filename in os.listdir(self.path):
                self.add_egg(filename)
        except:
            # Ignore missing directory unless it's needed to cache remote
            # packages. A remote basket is unusable without the local cache.
            if url:
                raise

        try: self.initialize()
        except: pass

    def _prepare_cache(self, url):
        import urllib

        path = os.path.join(self._cache_dir, urllib.quote(url, ''))
        if not os.path.isdir(path):
            os.makedirs(path)
        return path

    @classmethod
    def _is_egg(cls, filename):
        return filename[-4] == '.' \
                and filename[-3:].lower() == 'egg'

    def add_egg(self, egg, **metadata):
        if not self._is_egg(egg):
            return

        from pkg_resources import Distribution, EGG_DIST

        dist = Distribution.from_location(self.path + egg, egg)
        dist._transmute_basket = self
        dist._transmute_metadata = metadata

        # Prefer local packages.
        dist.precedence = EGG_DIST - 0.1

        self.distributions[egg] = dist

    def fill_environment(self, environment, requirements=None):
        for dist in self.distributions.itervalues():
            environment.add(dist)

        for req in requirements:
            project = req.project_name
            if project in self._projects:
                continue

            try: self.initialize_project(project)
            except: continue

            self._projects.add(project)

    def make_local(self, dist):
        if os.path.isfile(dist.location):
            return

        # Called from within catch-all in top-level require()
        self.fetch(dist, **dist._transmute_metadata)

    # Hooks for implementing custom baskets.
    #
    # These functions are called inside catch-all blocks. This is done both for
    # resilience against transient issues, network or otherwise, and to simplify
    # implementation of derived classes.
    #
    # Derived classes are not required to call the implementations herein.

    def initialize(self):
        """Instance initialization, called once at the end of __init__."""
        pass

    def initialize_project(self, project_name):
        """Called before fulfilling requirements, once per project_name."""
        pass

    def fetch(self, dist, **metadata):
        """Called from make_local if local copy does not exist."""
        raise RuntimeError('Unable to fetch: %s' % dist)


class PyPIBasket(Basket):
    """A proxy basket for eggs available in PyPI."""

    pypi_url = 'https://pypi.python.org/pypi'

    def fetch(self, dist, **metadata):
        import urllib2

        _download(urllib2.urlopen(metadata['url']),
                dist.location, metadata['md5_digest'])

    def initialize_project(self, project_name):
        import json
        import urllib2

        url = '%s/%s/json' % (self.pypi_url, project_name)
        req = urllib2.urlopen(url)
        metadata = json.load(req)

        for package in metadata['urls']:
            if not sys.version.startswith(package['python_version']) \
                    or package['packagetype'] != 'bdist_egg':
                continue
            self.add_egg(package['filename'], **package)


PYPI_BASKET = PyPIBasket(PyPIBasket.pypi_url)


def bootstrap():
    """Bootstrap 'transmute' package, making it available for import.

    Actual packages to be bootstrapped are defined in the global variable
    requirements.

    Latest packages are downloaded from PyPI, if available, and added to
    sys.path.
    """
    bootstrap_starting()

    import pkg_resources
    working_set = pkg_resources.WorkingSet([])

    try:
        require(working_set, [ PYPI_BASKET ], *requirements)
    except:
        raise
        bootstrap_failed()
    else:
        reset_system_path(working_set)
        bootstrap_succeeded()

def _clean_namespace():
    """Clean module's namespace.

    Used prior to a reload() of the module when the present module is used to
    bootstrap the Real Thing (tm).
    """
    global __doc__, os, sys
    del __doc__
    del os
    del sys

    global requirements, main, bootstrap_starting, bootstrap_succeeded, \
            bootstrap_failed
    del requirements
    del main
    del bootstrap_starting
    del bootstrap_succeeded
    del bootstrap_failed

    global _chunk_read, _md5, _copy, _download
    del _chunk_read
    del _md5
    del _copy
    del _download

    global reset_system_path, require, Basket, PyPIBasket, PYPI_BASKET
    del reset_system_path
    del require
    del Basket
    del PyPIBasket
    del PYPI_BASKET

    global bootstrap, _clean_namespace
    del bootstrap
    del _clean_namespace


if __name__ in requirements:
    bootstrap()

    # Module is filling in for a package so cleanup namespace and reload
    _clean_namespace()

    import sys
    reload(sys.modules[__name__])
else:
    import importlib
    try:
        for package in requirements:
            importlib.import_module(package)
    except ImportError:
        bootstrap()


# Running as self-contained script, invoke user hook.
if __name__ == '__main__':
    sys.exit(main())



#   Copyright 2014 Telenor Digital AS
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import re
from transmute.bootstrap import Basket

_basket_factory = {}
_basket = {}

_SCHEME_REGEX = re.compile("^[a-z][a-z0-9+.-]*$")
def register_basket_factory(scheme, factory):
    assert _SCHEME_REGEX.match(scheme)
    _basket_factory[scheme] = factory

def register_basket(url, basket):
    _basket[url] = basket

def _get_basket(url):
    scheme, colon, _ = url.partition(':')
    if colon \
            and scheme in _basket_factory:
        return _basket_factory[scheme](url)

    # Assume url is a local path
    return Basket(path=url)

def get_basket(url):
    if url not in _basket:
        _basket[url] = _get_basket(url)
    return _basket[url]



#   Copyright 2014 Telenor Digital AS
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import base64
import email.utils
import hashlib
import hmac
import json
import os
import urllib
import urllib2
import xml.etree.ElementTree

from transmute.basket import Basket, register_basket_factory
from transmute.bootstrap import _download


def _get_s3_endpoint():
    region = os.environ.get('AWS_DEFAULT_REGION') \
            or os.environ.get('EC2_REGION')
    if not region \
            or region == 'us-east-1':
        return 'https://s3.amazonaws.com'
    if region == 'EU':
        region = 'eu-west-1'
    return 'https://s3-%s.amazonaws.com' % region

def _get_aws_credentials():
    for provider in [
                _aws_credentials_from_file,
                _aws_credentials_from_environment,
                _aws_credentials_from_metadata,
            ]:
        try: return provider()
        except: pass

    return None, None, None

def _aws_credentials_from_file():
    config = {}
    with open(os.environ['AWS_CREDENTIAL_FILE']) as credential_file:
        for line in credential_file:
            name, sep, value = line.partition('=')
            if sep:
                config[name.strip()] = value.strip()
    return config['AWSAccessKeyId'], config['AWSSecretKey'], None

def _aws_credentials_from_environment():
    return os.environ['AWS_ACCESS_KEY'], os.environ['AWS_SECRET_KEY'], \
            os.environ.get('AWS_SECURITY_TOKEN', None)

def _aws_credentials_from_metadata():
    url = 'http://169.254.169.254/latest/meta-data/iam/security-credentials'
    role = urllib2.urlopen(url, timeout=1).read()
    credentials = json.load(urllib2.urlopen(url + '/' + role, timeout=1))
    return credentials['AccessKeyId'], credentials['SecretAccessKey'], \
            credentials['Token']


class _S3BucketFolder:
    """A read-only, view over a flat directory in AWS S3."""

    endpoint = _get_s3_endpoint()
    access_key, secret_key, security_token = _get_aws_credentials()

    def __init__(self, bucket, prefix=''):
        self.bucket = bucket
        self.prefix = prefix + '/'

    def _request(self, path, query=None):
        headers = { 'Host': self.bucket }
        self._authenticate_request(path, headers)
        url = self.endpoint + path + (query or '')

        request = urllib2.Request(url, headers=headers)
        response = urllib2.urlopen(request)
        if response.getcode() == 200:
            return response

        raise RuntimeError('%s: %s' % (response.getcode(), response.read()))

    def _xml_request(self, path, query=None):
        return xml.etree.ElementTree.parse(self._request(path, query))

    def _authenticate_request(self, path, headers):
        # See http://docs.aws.amazon.com/AmazonS3/latest/dev/RESTAuthentication.html
        # Shortcuts taken liberally, this is not a full implementation.

        if not self.access_key:
            return

        date = email.utils.formatdate()
        headers['Date'] = date

        message = 'GET\n\n\n' + date + '\n'
        if self.security_token:
            headers['x-amz-security-token'] = self.token
            message += 'x-amz-security-token:' + self.token + '\n'
        message += '/' + self.bucket + path

        h = hmac.new(self.secret, message, hashlib.sha1)
        signature = base64.b64encode(h.digest())

        headers['Authorization'] = 'AWS %s:%s' % (self.key, signature)

    def get_bucket_location(self):
        return self._xml_request('/?location').getroot().text

    def list_objects(self):
        """List objects in directory.

        Sub-directories are not listed or traversed. Entries don't contain the
        common S3 key prefix.
        """
        marker = ''
        query = '?delimiter=/&encoding-type=url&prefix=' \
                + urllib.quote_plus(self.prefix, '/')

        while True:
            result = self._xml_request('/', query=query + marker)
            for key in result.iterfind(
                    '{http://s3.amazonaws.com/doc/2006-03-01/}Contents/'
                    '{http://s3.amazonaws.com/doc/2006-03-01/}Key'):

                key = urllib.unquote_plus(key.text)
                yield key[len(self.prefix):]

            marker = result.find(
                    '{http://s3.amazonaws.com/doc/2006-03-01/}NextMarker')
            if marker is None:
                break
            marker = '&marker=' + marker.text

    def get_object(self, name):
        """Read object from S3.

        Returns a tuple consisting of the MD5 hash of the content and a
        file-like stream for it.
        """
        path = urllib.quote_plus('/' + self.prefix + name, '/')
        response = self._request(path)

        return response.headers['ETag'][1:-1], response


class S3Basket(Basket):
    def initialize(self):
        assert self.url.startswith('s3://')

        bucket, _, prefix = self.url[5:].partition('/')
        self.s3_bucket = _S3BucketFolder(bucket, prefix)

        for filename in self.s3_bucket.list_objects():
            self.add_egg(filename, filename=filename)

    def fetch(self, dist, **metadata):
        md5sum, data = self.s3_bucket.get_object(metadata['filename'])
        _download(data, dist.location, md5sum[1:-1])


register_basket_factory('s3', S3Basket)



#   Copyright 2014 Telenor Digital AS
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import pkg_resources
import transmute.basket
import transmute.bootstrap

class Resolver:
    """Find and manage lists of updated packages."""

    def __init__(self, requirements=None, sources=None):
        """Initialize a new Resolver object.

        requirements: string or list of strings listing package requirements.
        """
        self.baskets = []
        self.reset()

        if sources:
            self.add_source(*sources)
        if requirements:
            self.require(requirements)

    def reset(self):
        self.working_set = pkg_resources.WorkingSet([])

    @classmethod
    def _get_basket(cls, source):
        if isinstance(source, basestring):
            return transmute.basket.get_basket(source)
        return source

    @classmethod
    def _get_baskets(cls, *sources):
        return [ cls._get_basket(s) for s in sources ]

    def add_source(self, *sources):
        self.extend(self._get_baskets(sources))

    def require(self, requirements, sources=None):
        baskets = self.baskets
        if sources:
            # Make a copy
            baskets = baskets + self._get_baskets(*sources)

        transmute.bootstrap.require(self.working_set, baskets, *requirements)



#   Copyright 2014 Telenor Digital AS
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

try: from transmute._version import __version__
except ImportError: __version__ = 'unknown'

import transmute.basket
from transmute.bootstrap import PYPI_BASKET
from transmute.resolver import Resolver
from transmute.transmuter import Transmuter

PYPI_SOURCE = PYPI_BASKET
transmute.basket.register_basket(PYPI_BASKET.pypi_url, PYPI_BASKET)

_resolver = Resolver()
add_source = _resolver.add_source
require = _resolver.require
reset = _resolver.reset

def update(resolver=None):
    if resolver is None:
        resolver = globals()['_resolver']
    tm = Transmuter(resolver.working_set)
    tm.transmute()



# -*- coding: utf-8 -*-

from setuptools import Command, find_packages, setup
from setuptools.command.egg_info import egg_info

import datetime
import os.path
import subprocess


README = 'README.md'
VERSION_FILE = 'transmute/_version.py'


# We use README.md for GitHub's sake, generate README.rst for PyPI.
#
try:
    # Requires pandoc, the tool, and pypandoc, the python module
    import pypandoc
    long_description = pypandoc.convert(README, 'rst')
except:
    with open(README) as readme:
        long_description = readme.read()
else:
    with open('README.rst', 'w') as rst:
        rst.write(long_description)


# Generate project version number from git tags
#
try: execfile(VERSION_FILE)
except: __version__ = 'unknown'

_git_revision = None
def get_git_revision():
    """Fetch git revision for use as the package version."""

    global _git_revision
    if _git_revision is None:
        wd = os.path.dirname(__file__) or None
        revision = subprocess.check_output(
                [ 'git', 'describe', '--always', '--dirty' ], cwd=wd)
        _git_revision = revision.rstrip()
    return _git_revision

def update_version_file():
    version_file_content = '\n'.join([
        "# File generated by %s on %s" % (__file__, datetime.date.today()),
        "__version__ = '%s'" % get_git_revision(),
        "" # End of file line-break
    ])
    with open(VERSION_FILE, 'w') as version_file:
        version_file.write(version_file_content)


class UpdateVersion(Command):

    description = "update package version information from version control"
    user_options = []

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        git_revision = get_git_revision()
        if git_revision != __version__:
            update_version_file()
            print 'New package version: %s (was: %s)' \
                    % (git_revision, __version__)
            self.distribution.metadata.version = git_revision


class EggInfo(egg_info):
    def finalize_options(self):
        # Must update version before egg_info sets internal egg_version in
        # finalize_options.
        self.run_command('update_version')
        egg_info.finalize_options(self)


# Alright, let's make this setuptools thingie run!
#
project_metadata = {
    'name':         'transmute',
    'version':      __version__,
    'author':       'João Abecasis',
    'author_email': 'joao@comoyo.com',
    'url':          'https://github.com/comoyo/python-transmute',
    'description':  'Automatically update Python Eggs on application startup.',
    'classifiers':  [
                        "Programming Language :: Python",
                        "License :: OSI Approved :: Apache Software License",
                        "Operating System :: OS Independent",
                        "Development Status :: 3 - Alpha",
                        "Intended Audience :: Developers",
                        "Topic :: Software Development :: Libraries :: Python Modules",
                        "Topic :: System :: Installation/Setup",
                        "Topic :: System :: Software Distribution",
                    ],
    'long_description': long_description,
    'packages':     find_packages(exclude=[ 'tests*' ]),
    'cmdclass':     {
                        'egg_info': EggInfo,
                        'update_version': UpdateVersion,
                    },
}
# Go!
setup(**project_metadata)
